"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse

def get_argparser():
	parser = argparse.ArgumentParser(description="""Generate an attestation keypair, using the provided Owner private
key to sign the Owner Certificate.""")
	parser.add_argument("--key", help="Which endorsement scheme to use (1 or 2)", type=auto_int)
	parser.add_argument("--certificate", help="""Optional certificate to store if finalizing the endorsement (hex
encoded), if no private key is specified""")
	parser.add_argument("--privateKey", help="""Optional private key to use to create a test certificate (hex encoded),
if no certificate is specified""")
	parser.add_argument("--targetId", help="The device's target ID (default is Ledger Blue)", type=auto_int)
	parser.add_argument("--issuerKey", help="Issuer key (hex encoded, default is batch 1)")
	parser.add_argument("--rootPrivateKey", help="SCP Host private key")
	parser.add_argument("--apdu", help="Display APDU log", action='store_true')
	return parser

def auto_int(x):
		return int(x, 0)

def hexstr(bstr):
	if (sys.version_info.major == 3):
		return binascii.hexlify(bstr).decode()
	if (sys.version_info.major == 2):
		return binascii.hexlify(bstr)
	return ""

def getDeployedSecretV2(dongle, masterPrivate, targetid, issuerKey):
		testMaster = PrivateKey(bytes(masterPrivate))
		testMasterPublic = bytearray(testMaster.pubkey.serialize(compressed=False))
		targetid = bytearray(struct.pack('>I', targetid))

		# identify
		apdu = bytearray([0xe0, 0x04, 0x00, 0x00]) + bytearray([len(targetid)]) + targetid
		dongle.exchange(apdu)

		# walk the chain
		nonce = os.urandom(8)
		apdu = bytearray([0xe0, 0x50, 0x00, 0x00]) + bytearray([len(nonce)]) + nonce
		auth_info = dongle.exchange(apdu)
		batch_signer_serial = auth_info[0:4]
		deviceNonce = auth_info[4:12]

		# if not found, get another pair
		#if cardKey != testMasterPublic:
		#	   raise Exception("Invalid batch public key")

		dataToSign = bytes(bytearray([0x01]) + testMasterPublic)
		signature = testMaster.ecdsa_sign(bytes(dataToSign))
		signature = testMaster.ecdsa_serialize(signature)
		certificate = bytearray([len(testMasterPublic)]) + testMasterPublic + bytearray([len(signature)]) + signature
		apdu = bytearray([0xE0, 0x51, 0x00, 0x00]) + bytearray([len(certificate)]) + certificate
		dongle.exchange(apdu)

		# provide the ephemeral certificate
		ephemeralPrivate = PrivateKey()
		ephemeralPublic = bytearray(ephemeralPrivate.pubkey.serialize(compressed=False))
		dataToSign = bytes(bytearray([0x11]) + nonce + deviceNonce + ephemeralPublic)
		signature = testMaster.ecdsa_sign(bytes(dataToSign))
		signature = testMaster.ecdsa_serialize(signature)
		certificate = bytearray([len(ephemeralPublic)]) + ephemeralPublic + bytearray([len(signature)]) + signature
		apdu = bytearray([0xE0, 0x51, 0x80, 0x00]) + bytearray([len(certificate)]) + certificate
		dongle.exchange(apdu)

		# walk the device certificates to retrieve the public key to use for authentication
		index = 0
		last_pub_key = PublicKey(binascii.unhexlify(issuerKey), raw=True)
		device_pub_key = None
		while True:
				if index == 0:
						certificate = bytearray(dongle.exchange(bytearray.fromhex('E052000000')))
				elif index == 1:
						certificate = bytearray(dongle.exchange(bytearray.fromhex('E052800000')))
				else:
						break
				if len(certificate) == 0:
						break
				offset = 1
				certificateHeader = certificate[offset : offset + certificate[offset-1]]
				offset += certificate[offset-1] + 1
				certificatePublicKey = certificate[offset : offset + certificate[offset-1]]
				offset += certificate[offset-1] + 1
				certificateSignatureArray = certificate[offset : offset + certificate[offset-1]]
				certificateSignature = last_pub_key.ecdsa_deserialize(bytes(certificateSignatureArray))
				# first cert contains a header field which holds the certificate's public key role
				if index == 0:
						certificateSignedData = bytearray([0x02]) + certificateHeader + certificatePublicKey
						# Could check if the device certificate is signed by the issuer public key
				# ephemeral key certificate
				else:
						certificateSignedData = bytearray([0x12]) + deviceNonce + nonce + certificatePublicKey
				if not last_pub_key.ecdsa_verify(bytes(certificateSignedData), certificateSignature):
					print("pub key not signed by last_pub_key")
					return None
				last_pub_key = PublicKey(bytes(certificatePublicKey), raw=True)
				if index == 0:
					device_pub_key = last_pub_key
				index = index + 1

		return device_pub_key

if __name__ == '__main__':
	from .comm import getDongle
	from .ecWrapper import PrivateKey, PublicKey
	import hashlib
	import struct
	import os
	import sys
	import binascii

	args = get_argparser().parse_args()

	if args.key == None:
			raise Exception("Missing endorsement scheme number")
	if args.key != 1 and args.key != 2:
			raise Exception("Invalid endorsement scheme number")
	if args.targetId == None:
			args.targetId = 0x31000002
	if args.issuerKey == None:
			args.issuerKey = "0490f5c9d15a0134bb019d2afd0bf297149738459706e7ac5be4abc350a1f818057224fce12ec9a65de18ec34d6e8c24db927835ea1692b14c32e9836a75dad609"
	if (args.privateKey != None) and (args.certificate != None):
		raise Exception("Cannot specify both certificate and privateKey")

	privateKey = PrivateKey()
	publicKey = str(privateKey.pubkey.serialize(compressed=False))
	if not args.rootPrivateKey:
		args.rootPrivateKey = privateKey.serialize()
	else:
		# parse the issuer key from the rootprivate key
		issuerPrivKey = PrivateKey(bytes(bytearray.fromhex(args.rootPrivateKey)))
		args.issuerKey = binascii.hexlify(issuerPrivKey.pubkey.serialize(compressed=False)).decode('utf-8')

	dongle = getDongle(args.apdu)
	print(args.rootPrivateKey)
	publicKey = getDeployedSecretV2(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId, args.issuerKey)

	if args.certificate == None:
			apdu = bytearray([0xe0, 0xC0, args.key, 0x00, 0x00])
			response = dongle.exchange(apdu)
			print("Public key " + hexstr(response[0:65]))
			m = hashlib.sha256()
			m.update(bytes(b"\xff")) # Endorsement role
			m.update(bytes(response[0:65]))
			digest = m.digest()
			signature = publicKey.ecdsa_deserialize(bytes(response[65:]))
			if not publicKey.ecdsa_verify(bytes(digest), signature, raw=True):
				raise Exception("Issuer certificate not verified")
			if args.privateKey != None:
				privateKey = PrivateKey(bytes(bytearray.fromhex(args.privateKey)))
				dataToSign = bytes(bytearray([0xfe]) + response[0:65])
				signature = privateKey.ecdsa_sign(bytes(dataToSign))
				args.certificate = hexstr(privateKey.ecdsa_serialize(signature))

	if args.certificate != None:
			certificate = bytearray.fromhex(args.certificate)
			apdu = bytearray([0xe0, 0xC2, 0x00, 0x00, len(certificate)]) + certificate
			dongle.exchange(apdu)
			print("Endorsement setup finalized")



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse

def get_argparser():
    parser = argparse.ArgumentParser(description="""Use attestation to determine if the device is a genuine Ledger
device.""")
    parser.add_argument("--targetId", help="The device's target ID (default is Ledger Blue)", type=auto_int)
    parser.add_argument("--issuerKey", help="Issuer key (hex encoded, default is batch 1)")
    parser.add_argument("--apdu", help="Display APDU log", action='store_true')
    return parser

def auto_int(x):
    return int(x, 0)

def getDeployedSecretV2(dongle, masterPrivate, targetId, issuerKey):
    testMaster = PrivateKey(bytes(masterPrivate))
    testMasterPublic = bytearray(testMaster.pubkey.serialize(compressed=False))
    targetid = bytearray(struct.pack('>I', targetId))

    # identify
    apdu = bytearray([0xe0, 0x04, 0x00, 0x00]) + bytearray([len(targetid)]) + targetid
    dongle.exchange(apdu)

    # walk the chain
    nonce = os.urandom(8)
    apdu = bytearray([0xe0, 0x50, 0x00, 0x00]) + bytearray([len(nonce)]) + nonce
    auth_info = dongle.exchange(apdu)
    batch_signer_serial = auth_info[0:4]
    deviceNonce = auth_info[4:12]

    # if not found, get another pair
    #if cardKey != testMasterPublic:
    #       raise Exception("Invalid batch public key")

    dataToSign = bytes(bytearray([0x01]) + testMasterPublic)
    signature = testMaster.ecdsa_sign(bytes(dataToSign))
    signature = testMaster.ecdsa_serialize(signature)
    certificate = bytearray([len(testMasterPublic)]) + testMasterPublic + bytearray([len(signature)]) + signature
    apdu = bytearray([0xE0, 0x51, 0x00, 0x00]) + bytearray([len(certificate)]) + certificate
    dongle.exchange(apdu)

    # provide the ephemeral certificate
    ephemeralPrivate = PrivateKey()
    ephemeralPublic = bytearray(ephemeralPrivate.pubkey.serialize(compressed=False))
    dataToSign = bytes(bytearray([0x11]) + nonce + deviceNonce + ephemeralPublic)
    signature = testMaster.ecdsa_sign(bytes(dataToSign))
    signature = testMaster.ecdsa_serialize(signature)
    certificate = bytearray([len(ephemeralPublic)]) + ephemeralPublic + bytearray([len(signature)]) + signature
    apdu = bytearray([0xE0, 0x51, 0x80, 0x00]) + bytearray([len(certificate)]) + certificate
    dongle.exchange(apdu)

    # walk the device certificates to retrieve the public key to use for authentication
    index = 0
    last_pub_key = PublicKey(binascii.unhexlify(issuerKey), raw=True)
    devicePublicKey = None
    while True:
        if index == 0:
            certificate = bytearray(dongle.exchange(bytearray.fromhex('E052000000')))
        elif index == 1:
            certificate = bytearray(dongle.exchange(bytearray.fromhex('E052800000')))
        else:
            break
        offset = 1
        certificateHeader = certificate[offset : offset + certificate[offset-1]]
        offset += certificate[offset-1] + 1
        certificatePublicKey = certificate[offset : offset + certificate[offset-1]]
        offset += certificate[offset-1] + 1
        certificateSignatureArray = certificate[offset : offset + certificate[offset-1]]
        certificateSignature = last_pub_key.ecdsa_deserialize(bytes(certificateSignatureArray))
        # first cert contains a header field which holds the certificate's public key role
        if index == 0:
            devicePublicKey = certificatePublicKey
            certificateSignedData = bytearray([0x02]) + certificateHeader + certificatePublicKey
            # Could check if the device certificate is signed by the issuer public key
        # ephemeral key certificate
        else:
            certificateSignedData = bytearray([0x12]) + deviceNonce + nonce + certificatePublicKey
        if not last_pub_key.ecdsa_verify(bytes(certificateSignedData), certificateSignature):
            return None
        last_pub_key = PublicKey(bytes(certificatePublicKey), raw=True)
        index = index + 1

    # Commit device ECDH channel
    dongle.exchange(bytearray.fromhex('E053000000'))
    secret = last_pub_key.ecdh(binascii.unhexlify(ephemeralPrivate.serialize()))
    if targetId&0xF == 0x2:
        return secret[0:16]
    elif targetId&0xF >= 0x3:
        ret = {}
        ret['ecdh_secret'] = secret
        ret['devicePublicKey'] = devicePublicKey
        return ret

if __name__ == '__main__':
    from .ecWrapper import PrivateKey, PublicKey
    from .comm import getDongle
    from .commException import CommException
    from .hexLoader import HexLoader
    import struct
    import os
    import binascii

    args = get_argparser().parse_args()

    if args.targetId == None:
        args.targetId = 0x31000002

    if args.issuerKey == None:
        args.issuerKey = "0490f5c9d15a0134bb019d2afd0bf297149738459706e7ac5be4abc350a1f818057224fce12ec9a65de18ec34d6e8c24db927835ea1692b14c32e9836a75dad609"

    privateKey = PrivateKey()
    publicKey = str(privateKey.pubkey.serialize(compressed=False))
    args.rootPrivateKey = privateKey.serialize()

    genuine = False
    ui = False
    customCA = False

    dongle = getDongle(args.apdu)
    version = None

    secret = getDeployedSecretV2(dongle, bytearray.fromhex(args.rootPrivateKey), args.targetId, args.issuerKey)
    if secret != None:
        try:
            loader = HexLoader(dongle, 0xe0, True, secret)
            version = loader.getVersion()
            genuine = True
            apps = loader.listApp()
            while len(apps) != 0:
                for app in apps:
                    if (app['flags'] & 0x08):
                        ui = True
                    if (app['flags'] & 0x400):
                        customCA = True
                apps = loader.listApp(False)
        except:
            genuine = False        
    if genuine:
        if ui:
            print ("WARNING : Product is genuine but has a UI application loaded")
        if customCA:
            print ("WARNING : Product is genuine but has a Custom CA loaded")
        if not ui and not customCA:
            print ("Product is genuine")
        print ("SE Version " + version['osVersion'])
        print ("MCU Version " + version['mcuVersion'])
        if 'mcuHash' in version:
            print ("MCU Hash " + binascii.hexlify(version['mcuHash']).decode('ascii'))
    else:
        print ("Product is NOT genuine")



"""
*******************************************************************************
*   Ledger Blue
*   (c) 2016 Ledger
*
*  Licensed under the Apache License, Version 2.0 (the "License");
*  you may not use this file except in compliance with the License.
*  You may obtain a copy of the License at
*
*      http://www.apache.org/licenses/LICENSE-2.0
*
*  Unless required by applicable law or agreed to in writing, software
*  distributed under the License is distributed on an "AS IS" BASIS,
*  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
*  See the License for the specific language governing permissions and
*  limitations under the License.
********************************************************************************
"""

import argparse

def get_argparser():
	parser = argparse.ArgumentParser(description="""Read a sequence of command APDUs from a file and send them to the
device. The file must be formatted as hex, with one CAPDU per line.""")
	parser.add_argument("--fileName", help="The name of the APDU script to load")
	parser.add_argument("--apdu", help="Display APDU log", action='store_true')
	parser.add_argument("--scp", help="Open a Secure Channel to exchange APDU", action='store_true')
	parser.add_argument("--targetId", help="The device's target ID (default is Ledger Nano S)", type=auto_int)
	parser.add_argument("--rootPrivateKey", help="""The Signer private key used to establish a Secure Channel (otherwise
a random one will be generated)""")
	return parser

def hexstr(bstr):
	if (sys.version_info.major == 3):
		return binascii.hexlify(bstr).decode()
	if (sys.version_info.major == 2):
		return binascii.hexlify(bstr)
	return ""

def auto_int(x):
	return int(x, 0)

if __name__ == '__main__':
	from .comm import getDongle
	from .deployed import getDeployedSecretV2
	from .ecWrapper import PrivateKey
	import sys
	import fileinput
	import binascii
	from .hexLoader import HexLoader

	args = get_argparser().parse_args()

	if args.targetId is None:
		args.targetId = 0x31100002
	if not args.fileName:
		#raise Exception("Missing fileName")
		file = sys.stdin
	else:
		file = open(args.fileName, "r")


	class SCP:

		def __init__(self, dongle, targetId, rootPrivateKey):
			secret = getDeployedSecretV2(dongle, rootPrivateKey, targetId)
			self.loader = HexLoader(dongle, 0xe0, True, secret)

		def encryptAES(self, data):
			return self.loader.scpWrap(data);

		def decryptAES(self, data):
			return self.loader.scpUnwrap(data);

	dongle = getDongle(args.apdu)
	if args.scp:
		if args.rootPrivateKey is None:
			privateKey = PrivateKey()
			publicKey = binascii.hexlify(privateKey.pubkey.serialize(compressed=False))
			print("Generated random root public key : %s" % publicKey)
			args.rootPrivateKey = privateKey.serialize()
			scp = SCP(dongle, args.targetId, bytearray.fromhex(args.rootPrivateKey))

	for data in file:
		data = binascii.unhexlify(data.replace("\n", ""))
		if len(data) < 5:
			continue
		if args.scp:
			apduData = data[4:]
			apduData = scp.encryptAES(bytes(apduData))
			apdu = bytearray([data[0], data[1], data[2], data[3], len(apduData)]) + bytearray(apduData)
			result = dongle.exchange(apdu)
			result = scp.decryptAES((result))
		else:
			result = dongle.exchange(bytearray(data))
		if args.apdu:
			print("<= Clear " + str(result))
